/* auth_logs.c: Prints out the authorisation logging info */

# ifndef lint
static char Rcsid[] = "@(#)$Header: /cs/research/pp/hubris/pp-beta/Src/submit/RCS/auth_logs.c,v 5.0 90/09/20 16:22:36 pp Exp Locker: pp $";
# endif

/*
 * $Header: /cs/research/pp/hubris/pp-beta/Src/submit/RCS/auth_logs.c,v 5.0 90/09/20 16:22:36 pp Exp Locker: pp $
 *
 * $Log:	auth_logs.c,v $
 * Revision 5.0  90/09/20  16:22:36  pp
 * rcsforce : 5.0 public release
 * 
 */



#include "util.h"
#include "q.h"
#include <isode/cmd_srch.h>


#define gval(x)                 ((x -> ad_type) == AD_X400_TYPE ? \
                                        (x -> ad_r400adr) : (x -> ad_r822adr))

/* -- externals -- */
extern CMD_TABLE	authtbl_statmsgs[];
extern CHAN             *ch_inbound;
extern ADDR             *ad_originator;
extern ADDR             *ad_recip;
extern char             *msg_unique;
extern char             *mgt_inhost;
extern int		auth_loglev;
extern long             msg_size;


/* -- local routines -- */
void			auth_log();
static int		auth_isdr();
static int		auth_isgood();
static void		gen_auth_logs();
static void		gen_auth_notok();
static void		gen_auth_ok();
static void		gen_auth_stats();
static void		gen_auth_warn();
static void		wr2statfile();
static void		wrdr2statfile();
static void		wrmsg2statfile();




/* -------------------  Begin  Routines  ------------------------------------ */




void auth_log()
{
        ADDR    *ad;

        PP_TRACE (("auth_log()"));

        for (ad = ad_recip; ad; ad = ad->ad_next)
                gen_auth_logs (ad);
}




/* ----------------------------- Logging ------------------------------------ */





static void gen_auth_logs (ad)
ADDR   *ad;
{
	LIST_RCHAN	*lp;
	int		retval = NOTOK;

	PP_TRACE (("gen_auth_logs(%s)", gval(ad)));

	for (lp = ad -> ad_outchan; lp != NULLIST_RCHAN; lp = lp -> li_next)
		if ((retval = auth_isgood (lp)) == OK)
			break; 

	gen_auth_stats (ad, retval);  	
	gen_auth_warn (ad, retval);
}




static void gen_auth_stats (ad, result)
ADDR	*ad;
int     result;
{

	PP_TRACE (("gen_auth_stats '%d' '%s'", result, gval (ad)));

	switch (result) {
	case OK: 
		gen_auth_ok (ad);
		break;
	default:
		gen_auth_notok (ad);
		break;
	}
}




static void gen_auth_ok (ad)
ADDR		*ad;
{
	LIST_RCHAN	*lp = ad -> ad_outchan;

	PP_TRACE (("gen_auth_ok '%s'", gval (ad)));

	switch (auth_loglev) {
	default:
	case AUTH_LOGLEVEL_LOW:
	case AUTH_LOGLEVEL_MEDIUM:
		for (; lp != NULLIST_RCHAN; lp = lp -> li_next)
			if (auth_isgood (lp) == OK) {
				wr2statfile (ad, lp);
				break;
			}
		return;

	case AUTH_LOGLEVEL_HIGH:
		for (; lp != NULLIST_RCHAN; lp = lp -> li_next)
			wr2statfile (ad, lp);
		return;
	}
}





static void gen_auth_notok (ad)
ADDR		*ad;
{
	LIST_RCHAN	*lp = ad -> ad_outchan;

	PP_TRACE (("gen_auth_notok '%s'", gval (ad)));

	switch (auth_loglev) {
	default:
	case AUTH_LOGLEVEL_LOW:
		if (lp != NULLIST_RCHAN)
			wr2statfile (ad, lp);	
		return;

	case AUTH_LOGLEVEL_MEDIUM:
	case AUTH_LOGLEVEL_HIGH:
		for (; lp != NULLIST_RCHAN; lp = lp -> li_next)
			wr2statfile (ad, lp);
		return;
	}
}




static int auth_isgood (lp)
LIST_RCHAN	*lp;
{
	if (lp == NULLIST_RCHAN)
		return FALSE; 

	switch (lp -> li_auth -> status) {
	case AUTH_OK:
	case AUTH_DR_OK:
	case AUTH_FAILED_TEST: 
		return OK;
	default:
		return NOTOK;
	}
}




static int auth_isdr (lp)
LIST_RCHAN	*lp;
{
	if (lp == NULLIST_RCHAN)
		return FALSE; 

	switch (lp -> li_auth -> status) {
	case AUTH_DR_OK:
	case AUTH_DR_FAILED:
		return OK;
	}
	return NOTOK;
}




static void wr2statfile (ad, lp)
ADDR		*ad;
LIST_RCHAN	*lp;
{

	if (auth_isdr (lp) == OK)
		wrdr2statfile (ad, lp);
	else 
		wrmsg2statfile (ad, lp);
}




static void wrmsg2statfile (ad, lp)
ADDR		*ad;
LIST_RCHAN	*lp;
{
	AUTH	*au = lp -> li_auth;
	char	*cp;
	char 	*argv[100];
	int	argc;
	char	buffer[BUFSIZ];
	char	sizebuf[30];

	PP_TRACE (("submit/wrmsg2statfile ('%s')", au -> reason));


	cp = rcmd_srch (au -> status, authtbl_statmsgs);

	if (cp == NULLCP)
		cp = rcmd_srch (AUTH_FAILED, authtbl_statmsgs);

	argc = 0;
	argv[argc++] = cp;
	argv[argc++] = msg_unique;
	argv[argc++] = ch_inbound -> ch_name;
	argv[argc++] = "->";
	argv[argc++] = au -> chan -> li_chan -> ch_name;
	argv[argc++] = "=";
	argv[argc++] = "size";
	(void) sprintf (sizebuf, "%d", msg_size);
	argv[argc++] = sizebuf;
	argv[argc++] = "=";
	argv[argc++] = "sender";
	argv[argc++] = gval (ad_originator);
	argv[argc++] = mgt_inhost;
	argv[argc++] = "=";
	argv[argc++] = "recip";
	argv[argc++] = gval (ad);
	argv[argc++] = au -> mta -> li_mta;
	argv[argc++] = au -> reason;

	argv[argc] = NULLCP;

	arg2vstr (0, sizeof buffer, buffer, argv);
	PP_STAT (("%s", buffer));
}




static void wrdr2statfile (ad, lp)
ADDR		*ad;
LIST_RCHAN	*lp;
{
	AUTH	*au = lp -> li_auth;
	char	*cp;
	char	*argv[100];
	int	argc;
	char	buffer[BUFSIZ];
	char	sizebuf[30];

	PP_TRACE (("submit/wrdr2statfile ('%s')", au -> reason));


	cp = rcmd_srch (au -> status, authtbl_statmsgs);

	if (cp == NULLCP)
		cp = rcmd_srch (AUTH_DR_FAILED, authtbl_statmsgs);

	argc = 0;
	argv[argc++] = cp;
	argv[argc++] = msg_unique;
	argv[argc++] = ch_inbound -> ch_name;
	argv[argc++] = "=";
	argv[argc++] = "size";
	(void) sprintf (sizebuf, "%d", msg_size);
	argv[argc++] = sizebuf;
	argv[argc++] = "=";
	argv[argc++] = "recip";
	argv[argc++] = gval (ad);
	argv[argc++] = mgt_inhost;
	argv[argc] = NULLCP;

	arg2vstr (0, sizeof buffer, buffer, argv);
	PP_STAT (("%s", buffer));
}

static void gen_auth_warn (ad, result)
ADDR	*ad;
int     result;
{
	int		found_s = FALSE, 
			found_r = FALSE;
	LIST_RCHAN 	*lp = ad -> ad_outchan;

	/* --- *** ---
	Not implemented yet -  instead generateis PP_OPER
	log messages to indicate what would have happened
	--- *** --- */


	if (result == OK)	return;

	for (; lp != NULLIST_RCHAN; lp = lp -> li_next) {

		if (lp -> li_auth -> chan -> warnsender) {
			if (found_s == FALSE) { 
				PP_OPER (NULLCP, ("%s '%s' %s '%s'",
 					"Warning message to sender",
					gval (ad_originator), 
					"(not generated) : file", 
					lp -> li_auth -> chan -> warnsender));
				found_s = TRUE;
			}
		}			

		if (lp -> li_auth -> chan -> warnrecipient) {
			if (found_r == FALSE) {
				PP_OPER (NULLCP, ("%s '%s' %s '%s'",
 					"Warning message to recipient",
					gval (ad_originator), 
					"(not generated) : file", 
					lp -> li_auth -> chan -> warnsender));
				found_r = TRUE;
			}
		}
	}

}
